# This file is part of thermotools.
#
# Copyright 2015 Computational Molecular Biology Group, Freie Universitaet Berlin (GER)
#
# thermotools is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

r"""
Python interface to the WHAM estimator's lowlevel functions.
"""

import numpy as _np
cimport numpy as _np

__all__ = ['iterate_fi', 'iterate_fk']

cdef extern from "_wham.h":
    void _iterate_fi(
        double *log_N_K, double *log_N_i, double *f_K, double *b_K_i,
        int n_therm_states, int n_markov_states, double *scratch_M, double *scratch_T, double *f_i)
    void _iterate_fk(
        double *f_i, double *b_K_i, int n_therm_states, int n_markov_states,
        double *scratch_M, double *f_K)

def iterate_fi(
    _np.ndarray[double, ndim=1, mode="c"] log_N_K not None,
    _np.ndarray[double, ndim=1, mode="c"] log_N_i not None,
    _np.ndarray[double, ndim=1, mode="c"] f_K not None,
    _np.ndarray[double, ndim=2, mode="c"] b_K_i not None,
    _np.ndarray[double, ndim=1, mode="c"] scratch_M not None,
    _np.ndarray[double, ndim=1, mode="c"] scratch_T not None,
    _np.ndarray[double, ndim=1, mode="c"] f_i not None):
    r"""
    Calculate the reduced free energies f_i
        
    Parameters
    ----------
    log_N_K : numpy.ndarray(shape=(T), dtype=numpy.float64)
        log of the state counts in each of the T thermodynamic states
    log_N_i : numpy.ndarray(shape=(M), dtype=numpy.float64)
        log of the state counts in each of the M markov states
    f_K : numpy.ndarray(shape=(T), dtype=numpy.float64)
        reduced free energies of the T thermodynamic states
    b_K_i : numpy.ndarray(shape=(T, M), dtype=numpy.float64)
        bias energies in the T thermodynamic and M discrete Markov states
    scratch_M : numpy.ndarray(shape=(M), dtype=numpy.float64)
        scratch array
    scratch_T : numpy.ndarray(shape=(T), dtype=numpy.float64)
        scratch array
    f_K : numpy.ndarray(shape=(T), dtype=numpy.float64)
        target array for the reduced free energies of the T thermodynamic states

    Notes
    -----
    The iterate_fi() function computes

    .. math:
        f_i = -\ln\left( \frac{
                \sum_{K=0}^{N_T-1} N_i^{(K)}
            }{
                \sum_{K=0}^{N_T-1} \exp(f^{(K)}-b_i^{(K)}) \sum_{j=0}^{N_M-1} N_j^{(K)}
            }\right)

    which equals to

    .. math:
        f_i = \ln\left(
                \sum_{K=0}^{N_T-1} \exp\left(
                    f^{(K)} - b_i^{(K)} + \ln\left( \sum_{j=0}^{N_M-1} N_j^{(K)} \right)
                \right)
            \right) - \ln\left( \sum_{K=0}^{N_T-1} N_i^{(K)} \right)

    in the logsumexp scheme. Afterwards, we apply

    .. math:
        f_i \leftarrow  f_i + \ln\left(
                \sum_{j=0}^{N_M-1} \exp(-f_j)
            \right)

    which is equilvalent to a renormalisation of the stationary distribution.
    """
    _iterate_fi(
        <double*> _np.PyArray_DATA(log_N_K),
        <double*> _np.PyArray_DATA(log_N_i),
        <double*> _np.PyArray_DATA(f_K),
        <double*> _np.PyArray_DATA(b_K_i),
        b_K_i.shape[0],
        b_K_i.shape[1],
        <double*> _np.PyArray_DATA(scratch_M),
        <double*> _np.PyArray_DATA(scratch_T),
        <double*> _np.PyArray_DATA(f_i))

def iterate_fk(
    _np.ndarray[double, ndim=1, mode="c"] f_i not None,
    _np.ndarray[double, ndim=2, mode="c"] b_K_i not None,
    _np.ndarray[double, ndim=1, mode="c"] scratch_M not None,
    _np.ndarray[double, ndim=1, mode="c"] f_K not None):
    r"""
    Calculate the reduced thermodynamic free energies f_K
        
    Parameters
    ----------
    f_i : numpy.ndarray(shape=(M), dtype=numpy.float64)
        reduced free energies of the M markov states
    b_K_i : numpy.ndarray(shape=(T, M), dtype=numpy.float64)
        bias energies in the T thermodynamic and M discrete Markov states
    scratch_M : numpy.ndarray(shape=(M), dtype=numpy.float64)
        scratch array
    f_K : numpy.ndarray(shape=(T), dtype=numpy.float64)
        target array for the reduced free energies of the T thermodynamic states

    Notes
    -----
    The iterate_fk() function computes

    .. math:
        f^{(K)} = -\ln\left(
                \sum_{i=0}^{N_M-1} \exp(-b_i^{(K)}-f_i)
            \right)

    which is already in the logsumexp form.
    """
    _iterate_fk(
        <double*> _np.PyArray_DATA(f_i),
        <double*> _np.PyArray_DATA(b_K_i),
        b_K_i.shape[0],
        b_K_i.shape[1],
        <double*> _np.PyArray_DATA(scratch_M),
        <double*> _np.PyArray_DATA(f_K))



# This file is part of thermotools.
#
# Copyright 2015 Computational Molecular Biology Group, Freie Universitaet Berlin (GER)
#
# thermotools is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

r"""
Python interface to the dTRAM estimator's lowlevel functions.
"""

import numpy as _np
cimport numpy as _np

__all__ = [
    'set_lognu',
    'iterate_lognu',
    'iterate_fi',
    'get_pk',
    'get_p',
    'get_fk']

cdef extern from "_dtram.h":
    void _set_lognu(
        double *log_nu_K_i, int *C_K_ij, int n_therm_states, int n_markov_states)
    void _iterate_lognu(
        double *log_nu_K_i, double *b_K_i, double *f_i, int *C_K_ij, int n_therm_states,
        int n_markov_states, double *scratch_M, double *new_log_nu_K_i)
    void _iterate_fi(
        double *log_nu_K_i, double *b_K_i, double *f_i, int *C_K_ij, int n_therm_states,
        int n_markov_states, double *scratch_TM, double *scratch_M, double *new_f_i)
    void _get_p(
        double *log_nu_i, double *b_i, double *f_i, int *C_ij,
        int n_markov_states, double *scratch_M, double *p_ij)
    void _get_fk(
        double *b_K_i, double *f_i, int n_therm_states, int n_markov_states,
        double *scratch_M, double *f_K)

def set_lognu(
    _np.ndarray[double, ndim=2, mode="c"] log_nu_K_i not None,
    _np.ndarray[int, ndim=3, mode="c"] C_K_ij not None):
    r"""
    Set the logarithm of the Lagrangian multipliers with an initial guess based
    on the transition counts

    Parameters
    ----------
    log_nu_K_i : numpy.ndarray(shape=(T, M), dtype=numpy.float64)
        log of the Lagrangian multipliers (allocated but unset)
    C_K_ij : numpy.ndarray(shape=(T, M, M), dtype=numpy.intc)
        multistate count matrix
    """
    _set_lognu(
        <double*> _np.PyArray_DATA(log_nu_K_i),
        <int*> _np.PyArray_DATA(C_K_ij),
        log_nu_K_i.shape[0],
        log_nu_K_i.shape[1])

def iterate_lognu(
    _np.ndarray[double, ndim=2, mode="c"] log_nu_K_i not None,
    _np.ndarray[double, ndim=2, mode="c"] b_K_i not None,
    _np.ndarray[double, ndim=1, mode="c"] f_i not None,
    _np.ndarray[int, ndim=3, mode="c"] C_K_ij not None,
    _np.ndarray[double, ndim=1, mode="c"] scratch_M not None,
    _np.ndarray[double, ndim=2, mode="c"] new_log_nu_K_i not None):
    r"""
    Update the logarithms of the Lagrangian multipliers

    Parameters
    ----------
    log_nu_K_i : numpy.ndarray(shape=(T, M), dtype=numpy.float64)
        log of the Lagrangian multipliers
    b_K_i : numpy.ndarray(shape=(T, M), dtype=numpy.intc)
        reduced bias energies of the T thermodynamic and M markov states
    f_i : numpy.ndarray(shape=(M,), dtype=numpy.float64)
        reduced unbiased free energies
    C_K_ij : numpy.ndarray(shape=(T, M, M), dtype=numpy.intc)
        multistate count matrix
    scratch_M : numpy.ndarray(shape=(M,), dtype=numpy.float64)
        scratch array for logsumexp operations
    new_log_nu_K_i : numpy.ndarray(shape=(T, M), dtype=numpy.float64)
        target array for the log of the Lagrangian multipliers
    """
    _iterate_lognu(
        <double*> _np.PyArray_DATA(log_nu_K_i),
        <double*> _np.PyArray_DATA(b_K_i),
        <double*> _np.PyArray_DATA(f_i),
        <int*> _np.PyArray_DATA(C_K_ij),
        log_nu_K_i.shape[0],
        log_nu_K_i.shape[1],
        <double*> _np.PyArray_DATA(scratch_M),
        <double*> _np.PyArray_DATA(new_log_nu_K_i))

def iterate_fi(
    _np.ndarray[double, ndim=2, mode="c"] log_nu_K_i not None,
    _np.ndarray[double, ndim=2, mode="c"] b_K_i not None,
    _np.ndarray[double, ndim=1, mode="c"] f_i not None,
    _np.ndarray[int, ndim=3, mode="c"] C_K_ij not None,
    _np.ndarray[double, ndim=2, mode="c"] scratch_TM not None,
    _np.ndarray[double, ndim=1, mode="c"] scratch_M not None,
    _np.ndarray[double, ndim=1, mode="c"] new_f_i not None):
    r"""
    Update the reduced unbiased free energies

    Parameters
    ----------
    log_nu_K_i : numpy.ndarray(shape=(T, M), dtype=numpy.float64)
        log of the Lagrangian multipliers
    b_K_i : numpy.ndarray(shape=(T, M), dtype=numpy.intc)
        reduced bias energies of the T thermodynamic and M markov states
    f_i : numpy.ndarray(shape=(M,), dtype=numpy.float64)
        reduced unbiased free energies
    C_K_ij : numpy.ndarray(shape=(T, M, M), dtype=numpy.intc)
        multistate count matrix
    scratch_TM : numpy.ndarray(shape=(T, M), dtype=numpy.float64)
        scratch array for logsumexp operations
    scratch_M : numpy.ndarray(shape=(M,), dtype=numpy.float64)
        scratch array for logsumexp operations
    new_f_i : numpy.ndarray(shape=(M,), dtype=numpy.float64)
        target array for the reduced unbiased free energies
    """
    _iterate_fi(
        <double*> _np.PyArray_DATA(log_nu_K_i),
        <double*> _np.PyArray_DATA(b_K_i),
        <double*> _np.PyArray_DATA(f_i),
        <int*> _np.PyArray_DATA(C_K_ij),
        log_nu_K_i.shape[0],
        log_nu_K_i.shape[1],
        <double*> _np.PyArray_DATA(scratch_TM),
        <double*> _np.PyArray_DATA(scratch_M),
        <double*> _np.PyArray_DATA(new_f_i))

def get_pk(
    _np.ndarray[double, ndim=2, mode="c"] log_nu_K_i not None,
    _np.ndarray[double, ndim=2, mode="c"] b_K_i not None,
    _np.ndarray[double, ndim=1, mode="c"] f_i not None,
    _np.ndarray[int, ndim=3, mode="c"] C_K_ij not None,
    _np.ndarray[double, ndim=1, mode="c"] scratch_M not None):
    r"""
    Compute the transition matrices for all thermodynamic states

    Parameters
    ----------
    log_nu_K_i : numpy.ndarray(shape=(T, M), dtype=numpy.float64)
        log of the Lagrangian multipliers
    b_K_i : numpy.ndarray(shape=(T, M), dtype=numpy.intc)
        reduced bias energies of the T thermodynamic and M markov states
    f_i : numpy.ndarray(shape=(M,), dtype=numpy.float64)
        reduced unbiased free energies
    C_K_ij : numpy.ndarray(shape=(T, M, M), dtype=numpy.intc)
        multistate count matrix
    scratch_M : numpy.ndarray(shape=(M,), dtype=numpy.float64)
        scratch array for logsumexp operations

    Returns
    -------
    p_K_ij : numpy.ndarray(shape=(T, M, M), dtype=numpy.float64)
        transition matrices for all thermodynamic states
    """
    p_K_ij = _np.zeros(shape=(C_K_ij.shape[0], C_K_ij.shape[1], C_K_ij.shape[2]), dtype=_np.float64)
    for K in range(log_nu_K_i.shape[0]):
        p_K_ij[K, :, :] = get_p(log_nu_K_i, b_K_i, f_i, C_K_ij, scratch_M, K)[:, :]
    return p_K_ij

def get_p(
    _np.ndarray[double, ndim=2, mode="c"] log_nu_K_i not None,
    _np.ndarray[double, ndim=2, mode="c"] b_K_i not None,
    _np.ndarray[double, ndim=1, mode="c"] f_i not None,
    _np.ndarray[int, ndim=3, mode="c"] C_K_ij not None,
    _np.ndarray[double, ndim=1, mode="c"] scratch_M not None,
    therm_state):
    r"""
    Compute the transition matrices for all thermodynamic states

    Parameters
    ----------
    log_nu_K_i : numpy.ndarray(shape=(T, M), dtype=numpy.float64)
        log of the Lagrangian multipliers
    b_K_i : numpy.ndarray(shape=(T, M), dtype=numpy.intc)
        reduced bias energies of the T thermodynamic and M markov states
    f_i : numpy.ndarray(shape=(M,), dtype=numpy.float64)
        reduced unbiased free energies
    C_K_ij : numpy.ndarray(shape=(T, M, M), dtype=numpy.intc)
        multistate count matrix
    scratch_M : numpy.ndarray(shape=(M,), dtype=numpy.float64)
        scratch array for logsumexp operations
    therm_state : int
        target thermodynamic state

    Returns
    -------
    p_ij : numpy.ndarray(shape=(M, M), dtype=numpy.float64)
        transition matrix for the target thermodynamic state
    """
    p_ij = _np.zeros(shape=(f_i.shape[0], f_i.shape[0]), dtype=_np.float64)
    _get_p(
        <double*> _np.PyArray_DATA(_np.ascontiguousarray(log_nu_K_i[therm_state, :])),
        <double*> _np.PyArray_DATA(_np.ascontiguousarray(b_K_i[therm_state, :])),
        <double*> _np.PyArray_DATA(f_i),
        <int*> _np.PyArray_DATA(_np.ascontiguousarray(C_K_ij[therm_state, :, :])),
        f_i.shape[0],
        <double*> _np.PyArray_DATA(scratch_M),
        <double*> _np.PyArray_DATA(p_ij))
    return p_ij

def get_fk(
    _np.ndarray[double, ndim=2, mode="c"] b_K_i not None,
    _np.ndarray[double, ndim=1, mode="c"] f_i not None,
    _np.ndarray[double, ndim=1, mode="c"] scratch_M not None,
    _np.ndarray[double, ndim=1, mode="c"] f_K not None):
    r"""
    Compute the transition matrices for all thermodynamic states

    Parameters
    ----------
    b_K_i : numpy.ndarray(shape=(T, M), dtype=numpy.intc)
        reduced bias energies of the T thermodynamic and M markov states
    f_i : numpy.ndarray(shape=(M,), dtype=numpy.float64)
        reduced unbiased free energies
    scratch_M : numpy.ndarray(shape=(M,), dtype=numpy.float64)
        scratch array for logsumexp operations
    f_K : numpy.ndarray(shape=(T,), dtype=numpy.float64)
        target array for the reduced thermodynamic free energies
    """
    _get_fk(
        <double*> _np.PyArray_DATA(b_K_i),
        <double*> _np.PyArray_DATA(f_i),
        b_K_i.shape[0],
        b_K_i.shape[1],
        <double*> _np.PyArray_DATA(scratch_M),
        <double*> _np.PyArray_DATA(f_K))



# This file is part of thermotools.
#
# Copyright 2015 Computational Molecular Biology Group, Freie Universitaet Berlin (GER)
#
# thermotools is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import thermotools.wham as wham
import thermotools.dtram as dtram
import numpy as np
from numpy.testing import assert_allclose

#   ************************************************************************************************
#   fixed-point iterations
#   ************************************************************************************************

def run_wham(N_K_i, b_K_i, maxiter, ftol):
    log_N_K = np.log(N_K_i.sum(axis=1))
    log_N_i = np.log(N_K_i.sum(axis=0))
    f_K = np.zeros(shape=b_K_i.shape[0], dtype=np.float64)
    f_i = np.zeros(shape=b_K_i.shape[1], dtype=np.float64)
    scratch_T = np.zeros(shape=f_K.shape, dtype=np.float64)
    scratch_M = np.zeros(shape=f_i.shape, dtype=np.float64)
    old_f_K = f_K.copy()
    stop = False
    for m in range(maxiter):
        wham.iterate_fk(f_i, b_K_i, scratch_M, f_K)
        nz = (old_f_K != 0.0)
        if (nz.sum() > 0) and (np.max(np.abs((f_K[nz] - old_f_K[nz])/old_f_K[nz])) < ftol):
            stop = True
        else:
            old_f_K[:] = f_K[:]
        wham.iterate_fi(log_N_K, log_N_i, f_K, b_K_i, scratch_M, scratch_T, f_i)
        if stop:
            break
    return f_K, f_i

def run_dtram(C_K_ij, b_K_i, maxiter, ftol):
    log_nu_K_i = np.zeros(shape=b_K_i.shape, dtype=np.float64)
    f_K = np.zeros(shape=b_K_i.shape[0], dtype=np.float64)
    f_i = np.zeros(shape=b_K_i.shape[1], dtype=np.float64)
    dtram.set_lognu(log_nu_K_i, C_K_ij)
    scratch_TM = np.zeros(shape=b_K_i.shape, dtype=np.float64)
    scratch_M = np.zeros(shape=f_i.shape, dtype=np.float64)
    old_f_K = f_K.copy()
    for m in range(maxiter):
        tmp_log_nu_K_i = np.copy(log_nu_K_i)
        dtram.iterate_lognu(tmp_log_nu_K_i, b_K_i, f_i, C_K_ij, scratch_M, log_nu_K_i)
        tmp_f_i = np.copy(f_i)
        dtram.iterate_fi(log_nu_K_i, b_K_i, tmp_f_i, C_K_ij, scratch_TM, scratch_M, f_i)
        dtram.get_fk(b_K_i, f_i, scratch_M, f_K)
        nz = (old_f_K != 0.0)
        if (nz.sum() > 0) and (np.max(np.abs((f_K[nz] - old_f_K[nz])/old_f_K[nz])) < ftol):
            break
        else:
            old_f_K[:] = f_K[:]
    return f_K, f_i, dtram.get_pk(log_nu_K_i, b_K_i, f_i, C_K_ij, scratch_M)

#   ************************************************************************************************
#   data generation functions
#   ************************************************************************************************

def tower_sample(distribution):
    cdf = np.cumsum(distribution)
    rnd = np.random.rand() * cdf[-1]
    ind = (cdf > rnd)
    idx = np.where(ind == True)
    return np.min(idx)

def draw_independent_samples(pi_K_i, n_samples):
    N_K_i = np.zeros(shape=pi_K_i.shape, dtype=np.intc)
    for K in range(pi_K_i.shape[0]):
        for s in range(n_samples):
            N_K_i[K, tower_sample(pi_K_i[K, :])] += 1
    return N_K_i

def draw_transition_counts(P_K_ij, n_samples, x0):
    """generates a discrete Markov chain"""
    C_K_ij = np.zeros(P_K_ij.shape, dtype=np.intc)
    for K in range(P_K_ij.shape[0]):
        x = x0
        for s in range(n_samples):
            x_new = tower_sample(P_K_ij[K, x, :])
            C_K_ij[K, x, x_new] += 1
            x = x_new
    return C_K_ij

#   ************************************************************************************************
#   test class
#   ************************************************************************************************

class TestThreeTwoModel(object):
    @classmethod
    def setup_class(cls):
        cls.energy = np.array([1.0, 2.0, 0.0], dtype=np.float64)
        cls.b_K_i = np.array([[0.0, 0.0, 0.0], 2.0 - cls.energy], dtype=np.float64)
        cls.pi_i = np.exp(-cls.energy) / np.exp(-cls.energy).sum()
        cls.f_i = -np.log(cls.pi_i)
        cls.Z_K = (np.exp(-cls.b_K_i) * cls.pi_i[np.newaxis, :]).sum(axis=1)
        cls.f_K = -np.log(cls.Z_K)
        cls.pi_K_i = (1.0 / cls.Z_K[:, np.newaxis]) * np.exp(-cls.b_K_i) * cls.pi_i[np.newaxis, :]
        cls.f_K_i = -np.log(cls.pi_K_i)
        metropolis = cls.energy[np.newaxis, :] - cls.energy[:, np.newaxis]
        metropolis[(metropolis < 0.0)] = 0.0
        selection = np.array([[0.5, 0.5, 0.0], [0.5, 0.0, 0.5], [0.0, 0.5, 0.5]], dtype=np.float64)
        metr_hast = selection * np.exp(-metropolis)
        for i in range(metr_hast.shape[0]):
            metr_hast[i, i] = 0.0
            metr_hast[i, i] = 1.0 - metr_hast[i, :].sum()
        cls.P_K_ij = np.array([metr_hast, selection])
        cls.n_samples = 10000
        cls.N_K_i = draw_independent_samples(cls.pi_K_i, cls.n_samples)
        cls.C_K_ij = draw_transition_counts(cls.P_K_ij, cls.n_samples, 0)
    @classmethod
    def teardown_class(cls):
        pass
    def setup(self):
        pass
    def teardown(self):
        pass
    def test_wham(self):
        f_K, f_i = run_wham(self.N_K_i, self.b_K_i, 50000, 1.0E-15)
        maxerr = 1.0E-1
        assert_allclose(f_K, self.f_K, atol=maxerr)
        assert_allclose(f_i, self.f_i, atol=maxerr)
    def test_dtram(self):
        f_K, f_i, P_K_ij = run_dtram(self.C_K_ij, self.b_K_i, 10000, 1.0E-15)
        maxerr = 1.0E-1
        assert_allclose(f_K, self.f_K, atol=maxerr)
        assert_allclose(f_i, self.f_i, atol=maxerr)
        assert_allclose(P_K_ij, self.P_K_ij, atol=maxerr)



# This file is part of thermotools.
#
# Copyright 2015 Computational Molecular Biology Group, Freie Universitaet Berlin (GER)
#
# thermotools is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from thermotools.dtram import iterate_lognu, iterate_fi, get_pk
import numpy as np
from numpy.testing import assert_allclose

def test_lognu_zero_counts():
    nm = 200
    nt = 100
    log_nu_K_i = np.zeros(shape=(nt, nm), dtype=np.float64)
    b_K_i = np.zeros(shape=(nt, nm), dtype=np.float64)
    f_i = np.zeros(shape=(nm,), dtype=np.float64)
    C_K_ij = np.zeros(shape=(nt, nm, nm), dtype=np.intc) # C_K_ii = 1.0E-10 (internal prior)
    scratch_i = np.zeros(shape=(nm,), dtype=np.float64)
    new_log_nu_K_i = np.zeros(shape=(nt, nm), dtype=np.float64)
    ref_log_nu_K_i = np.log(1.0E-10*np.ones(shape=(nt, nm), dtype=np.float64)) # (prior)
    iterate_lognu(log_nu_K_i, b_K_i, f_i, C_K_ij, scratch_i, new_log_nu_K_i)
    assert_allclose(new_log_nu_K_i, ref_log_nu_K_i, atol=1.0E-16)

def test_lognu_all_factors_unity():
    nm = 200
    nt = 100
    log_nu_K_i = np.zeros(shape=(nt, nm), dtype=np.float64)
    b_K_i = np.zeros(shape=(nt, nm), dtype=np.float64)
    f_i = np.zeros(shape=(nm,), dtype=np.float64)
    C_K_ij = np.ones(shape=(nt, nm, nm), dtype=np.intc)
    scratch_i = np.zeros(shape=(nm,), dtype=np.float64)
    new_log_nu_K_i = np.zeros(shape=(nt, nm), dtype=np.float64)
    ref_log_nu_K_i = np.log(nm*np.ones(shape=(nt, nm), dtype=np.float64))
    iterate_lognu(log_nu_K_i, b_K_i, f_i, C_K_ij, scratch_i, new_log_nu_K_i)
    assert_allclose(new_log_nu_K_i, ref_log_nu_K_i, atol=1.0E-16)

def test_lognu_K_range():
    nm = 200
    nt = 100
    log_nu_K_i = np.zeros(shape=(nt, nm), dtype=np.float64)
    for K in range(nt):
        log_nu_K_i[K, :] = np.log(K + 1.0)
    b_K_i = np.zeros(shape=(nt, nm), dtype=np.float64)
    f_i = np.zeros(shape=(nm,), dtype=np.float64)
    C_K_ij = np.ones(shape=(nt, nm, nm), dtype=np.intc)
    scratch_i = np.zeros(shape=(nm,), dtype=np.float64)
    new_log_nu_K_i = np.zeros(shape=(nt, nm), dtype=np.float64)
    ref_log_nu_K_i = np.log(nm*np.ones(shape=(nt, nm), dtype=np.float64))
    iterate_lognu(log_nu_K_i, b_K_i, f_i, C_K_ij, scratch_i, new_log_nu_K_i)
    assert_allclose(new_log_nu_K_i, ref_log_nu_K_i, atol=1.0E-16)


def test_fi_zero_counts():
    nm = 200
    nt = 100
    log_nu_K_i = np.zeros(shape=(nt, nm), dtype=np.float64)
    b_K_i = np.zeros(shape=(nt, nm), dtype=np.float64)
    f_i = np.zeros(shape=(nm,), dtype=np.float64)
    C_K_ij = np.zeros(shape=(nt, nm, nm), dtype=np.intc)
    scratch_TM = np.zeros(shape=(nt, nm), dtype=np.float64)
    scratch_M = np.zeros(shape=(nm,), dtype=np.float64)
    new_f_i = np.zeros(shape=(nm,), dtype=np.float64)
    ref_f_i = np.log(nm*np.ones(shape=(nm,), dtype=np.float64))
    iterate_fi(log_nu_K_i, b_K_i, f_i, C_K_ij, scratch_TM, scratch_M, new_f_i)
    assert_allclose(new_f_i, ref_f_i, atol=1.0E-16)

def test_fi_all_factors_unity():
    nm = 200
    nt = 100
    log_nu_K_i = np.zeros(shape=(nt, nm), dtype=np.float64)
    b_K_i = np.zeros(shape=(nt, nm), dtype=np.float64)
    f_i = np.zeros(shape=(nm,), dtype=np.float64)
    C_K_ij = np.ones(shape=(nt, nm, nm), dtype=np.intc)
    scratch_TM = np.zeros(shape=(nt, nm), dtype=np.float64)
    scratch_M = np.zeros(shape=(nm,), dtype=np.float64)
    new_f_i = np.zeros(shape=(nm,), dtype=np.float64)
    ref_f_i = np.log(nm*np.ones(shape=(nm,), dtype=np.float64))
    iterate_fi(log_nu_K_i, b_K_i, f_i, C_K_ij, scratch_TM, scratch_M, new_f_i)
    assert_allclose(new_f_i, ref_f_i, atol=1.0E-16)


def test_pij_zero_counts():
    nm = 200
    nt = 100
    log_nu_K_i = np.zeros(shape=(nt, nm), dtype=np.float64)
    b_K_i = np.zeros(shape=(nt, nm), dtype=np.float64)
    f_i = np.zeros(shape=(nm,), dtype=np.float64)
    C_K_ij = np.zeros(shape=(nt, nm, nm), dtype=np.intc)
    scratch_M = np.zeros(shape=(nm,), dtype=np.float64)
    p_K_ij = get_pk(log_nu_K_i, b_K_i, f_i, C_K_ij, scratch_M)
    ref_p_ij = np.eye(nm, dtype=np.float64)
    for K in range(nt):
        assert_allclose(p_K_ij[K, :, :], ref_p_ij, atol=1.0E-16)

def test_pij_all_factors_unity():
    nm = 200
    nt = 100
    log_nu_K_i = np.zeros(shape=(nt, nm), dtype=np.float64)
    b_K_i = np.zeros(shape=(nt, nm), dtype=np.float64)
    f_i = np.zeros(shape=(nm,), dtype=np.float64)
    C_K_ij = np.ones(shape=(nt, nm, nm), dtype=np.intc)
    scratch_M = np.zeros(shape=(nm,), dtype=np.float64)
    p_K_ij = get_pk(log_nu_K_i, b_K_i, f_i, C_K_ij, scratch_M)
    ref_p_ij = np.ones(shape=(nm, nm), dtype=np.float64) + np.eye(nm, dtype=np.float64)*1.0E-10
    ref_p_ij /= ref_p_ij.sum(axis=1)[:, np.newaxis]
    for K in range(nt):
        assert_allclose(p_K_ij[K, :, :], ref_p_ij, atol=1.0E-16)



# This file is part of thermotools.
#
# Copyright 2015 Computational Molecular Biology Group, Freie Universitaet Berlin (GER)
#
# thermotools is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

r"""
Python interface to the logsumexp summation scheme for numerically robust summation of exponentials.
"""

import numpy as _np
cimport numpy as _np

__all__ = ['logsumexp', 'logsumexp_pair']

cdef extern from "_lse.h":
    double _logsumexp(double *array, int length)
    double _logsumexp_pair(double a, double b)

def logsumexp(_np.ndarray[double, ndim=1, mode="c"] array not None):
    r"""
    Perform a summation of an array of exponentials via the logsumexp scheme
        
    Parameters
    ----------
    array : numpy.ndarray(dtype=numpy.float64)
        arguments of the exponentials

    Returns
    -------
    ln_sum : float
        logarithm of the sum of exponentials

    Notes
    -----
    The logsumexp() function returns

    .. math:
        \ln\left( \sum_{i=0}^{n-1} \exp(a_i) \right)

    where the :math:`a_i` are the :math:`n` values in the supplied array.
    """
    return _logsumexp(<double*> _np.PyArray_DATA(array), array.shape[0])

def logsumexp_pair(a, b):
    r"""
    Perform a summation of two exponentials via the logsumexp scheme
        
    Parameters
    ----------
    a : float
        arguments of the first exponential
    b : float
        arguments of the second exponential

    Returns
    -------
    ln_sum : float
        logarithm of the sum of exponentials

    Notes
    -----
    The logsumexp_pair() function returns

    .. math:
        \ln\left( \exp(a) + \exp(b) \right)

    where the :math:`a` and :math:`b` are the supplied values.
    """
    return _logsumexp_pair(a, b)



# This file is part of thermotools.
#
# Copyright 2015 Computational Molecular Biology Group, Freie Universitaet Berlin (GER)
#
# thermotools is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from nose.tools import assert_true
from thermotools.bar import df
import scipy.integrate
import numpy as np

def delta_f_gaussian():
    f1 = scipy.integrate.quad(make_gauss(N=100000, sigma=1, mu=0), -np.inf, np.inf)
    f2 = scipy.integrate.quad(make_gauss(N=100000, sigma=5, mu=0), -np.inf, np.inf)
    return np.log(f2[0]) - np.log(f1[0])

def make_gauss(N, sigma, mu):
    k = 1
    s = -1.0 / (2 * sigma**2)
    def f(x):
        return k * np.exp(s * (x - mu)**2)
    return f

def test_bar():
    x1 = np.random.normal(loc=0, scale=1.0, size=10000)
    x2 = np.random.normal(loc=0, scale=12.5, size=10050)
    u_x1_x1 = 0.5 * x1**2
    u_x2_x2 = 12.5 * x2**2
    u_x1_x2 = 0.5 * x2**2
    u_x2_x1 = 12.5 * x1**2
    dbIJ = u_x1_x1 - u_x2_x1
    dbJI = u_x2_x2 - u_x1_x2
    bar = df(dbIJ, dbJI, np.zeros(dbJI.shape[0]))
    assert_true(np.fabs(delta_f_gaussian() - bar) < 0.05)




#!/usr/bin/env python

# This file is part of thermotools.
#
# Copyright 2015 Computational Molecular Biology Group, Freie Universitaet Berlin (GER)
#
# thermotools is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

r"""
thermotools is a lowlevel implementation toolbox for the analyis of free energy calculations
"""

from setuptools import setup
from distutils.core import Extension
from sys import exit as sys_exit
import versioneer

try:
    from Cython.Distutils import build_ext
except ImportError:
    print("ERROR - please install the cython dependency manually:")
    print("pip install cython")
    sys_exit(1)

try:
    from numpy import get_include
except ImportError:
    print("ERROR - please install the numpy dependency manually:")
    print("pip install numpy")
    sys_exit(1)

ext_lse = Extension(
    "thermotools.lse",
    sources=["ext/lse/lse.pyx", "ext/lse/_lse.c"],
    include_dirs=[get_include()],
    extra_compile_args=["-O3"])
ext_bar = Extension(
    "thermotools.bar",
    sources=["ext/bar/bar.pyx", "ext/bar/_bar.c", "ext/lse/_lse.c"],
    include_dirs=[get_include()],
    extra_compile_args=["-O3"])
ext_wham = Extension(
    "thermotools.wham",
    sources=["ext/wham/wham.pyx", "ext/wham/_wham.c", "ext/lse/_lse.c"],
    include_dirs=[get_include()],
    extra_compile_args=["-O3"])
# ext_tram = Extension(
#     "thermotools.tram",
#     sources=["ext/tram/tram.pyx", "ext/tram/_tram.c", "ext/lse/_lse.c"],
#     include_dirs=[get_include()],
#     extra_compile_args=["-O3"])
ext_dtram = Extension(
    "thermotools.dtram",
    sources=["ext/dtram/dtram.pyx", "ext/dtram/_dtram.c", "ext/lse/_lse.c"],
    include_dirs=[get_include()],
    extra_compile_args=["-O3"])
# ext_xtram = Extension(
#     "thermotools.xtram",
#     sources=["ext/xtram/xtram.pyx", "ext/xtram/_xtram.c", "ext/lse/_lse.c"],
#     include_dirs=[get_include()],
#     extra_compile_args=["-O3"])

cmd_class = versioneer.get_cmdclass()
cmd_class.update({'build_ext': build_ext})

setup(
    cmdclass=cmd_class,
    ext_modules=[
        ext_lse,
        ext_bar,
        ext_wham,
        #ext_tram,
        ext_dtram,
        #ext_xtram,
        ],
    name='thermotools',
    version=versioneer.get_version(),
    description='Lowlevel implementation of free energy estimators',
    long_description='Lowlevel implementation of free energy estimators',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Console',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: GNU Lesser General Public License v3 or later (LGPLv3+)',
        'Natural Language :: English',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: POSIX',
        'Operating System :: Microsoft :: Windows',
        'Programming Language :: C',
        'Programming Language :: Cython',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Topic :: Scientific/Engineering :: Bio-Informatics',
        'Topic :: Scientific/Engineering :: Chemistry',
        'Topic :: Scientific/Engineering :: Mathematics',
        'Topic :: Scientific/Engineering :: Physics'],
    keywords=[
        'free energy',
        'Markov state model',
        'BAR',
        'WHAM',
        'TRAM',
        'dTRAM',
        'xTRAM'],
    url='https://github.com/markovmodel/thermotools',
    maintainer='Christoph Wehmeyer',
    maintainer_email='christoph.wehmeyer@fu-berlin.de',
    license='LGPLv3+',
    setup_requires=[
        'numpy>=1.7.1',
        'cython>=0.20',
        'setuptools>=0.6'],
    tests_require=['numpy>=1.7.1', 'nose>=1.3'],
    install_requires=['numpy>=1.7.1'],
    packages=['thermotools'],
    test_suite='nose.collector',
    scripts=[]
)



# This file is part of thermotools.
#
# Copyright 2015 Computational Molecular Biology Group, Freie Universitaet Berlin (GER)
#
# thermotools is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

r"""
Python interface to the BAR ratio initialisation
"""

import numpy as np
cimport numpy as np

__all__ = ['df']

cdef extern from "_bar.h":
    double _df(double *db_IJ, int L1, double *db_JI, int L2, double *scratch)

def df(np.ndarray[double, ndim=1, mode="c"] db_IJ not None,
       np.ndarray[double, ndim=1, mode="c"] db_JI not None,
       np.ndarray[double, ndim=1, mode="c"] scratch not None):
    
    """ Free energy differences between two thermodynamic states using Bennett's 
    acceptance ratio (BAR).
    Estimates the free energy difference between two thermodynamic states
    using Bennett's acceptance ratio (BAR) [1]_. As an input, we need
    a set of reduced bias energy differences. Reduced bias energy differences
    are given in units of the thermal energy, often denoted by 
    :math:`\Delta b^{IJ}(x) = (B^I(x \in J) - B(x \in I)) / kT^I`
    where B(x) is the bias energy function and kT is the thermal
    energy.
    Parameters
    ----------
    db_IJ : numpy.ndarray(shape=(L1,), dtype=np.float64)
        Reduced biased energy differences for samples generated in thermodynamic state I.
    db_JI : numpy.ndarray(shape=(L2,), dtype=np.float64)
        Reduced biased energy differences for samples generated in thermodynamic state J.
    sctatch : numpy.ndarray(shape=(max(L1,L2)), dtryp=np.float64)
        Empty scatch array for internal data processing
    Returns
    -------
    df : float
        free energy difference between states I and J defined by :math:`f^IJ = f^J-f^I`.
    References
    ----------
    .. [1] Bennett, C. H.: Efficient Estimation of Free Energy Differences from
        Monte Carlo Data. J. Comput. Phys. 22, 245-268 (1976)
    """
    return _df(
    	<double*> np.PyArray_DATA(db_IJ),
    	db_IJ.shape[0],
    	<double*> np.PyArray_DATA(db_JI), 
        db_JI.shape[0],
        <double*> np.PyArray_DATA(scratch))



# This file is part of thermotools.
#
# Copyright 2015 Computational Molecular Biology Group, Freie Universitaet Berlin (GER)
#
# thermotools is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from thermotools.wham import iterate_fi, iterate_fk
from thermotools.lse import logsumexp
import numpy as np
from numpy.testing import assert_allclose

def test_wham_fk_with_zeros():
    T = 5
    M = 10
    f_i = np.zeros(shape=(M,), dtype=np.float64)
    b_K_i = np.zeros(shape=(T, M), dtype=np.float64)
    scratch = np.zeros(shape=(M,), dtype=np.float64)
    f_K = np.zeros(shape=(T,), dtype=np.float64)
    ref = -np.log(M)
    iterate_fk(f_i, b_K_i, scratch, f_K)
    assert_allclose(f_K, ref, atol=1.0E-15)

def test_wham_fk_with_ascending_K():
    T = 5
    M = 10
    f_i = np.zeros(shape=(M,), dtype=np.float64)
    b_K_i = -np.log(np.array([[K]*M for K in range(1, T + 1)], dtype=np.float64)/float(M))
    scratch = np.zeros(shape=(M,), dtype=np.float64)
    f_K = np.zeros(shape=(T,), dtype=np.float64)
    ref = -np.log(np.arange(T).astype(np.float64) + 1)
    iterate_fk(f_i, b_K_i, scratch, f_K)
    assert_allclose(f_K, ref, atol=1.0E-15)

def test_wham_fi_with_zeros():
    T = 5
    M = 10
    N_K_i = np.ones(shape=(T, M), dtype=np.float64)
    log_N_K = np.log(N_K_i.sum(axis=1))
    log_N_i = np.log(N_K_i.sum(axis=0))
    f_i = np.zeros(shape=(M,), dtype=np.float64)
    b_K_i = np.zeros(shape=(T, M), dtype=np.float64)
    scratch_M = np.zeros(shape=(M,), dtype=np.float64)
    scratch_T = np.zeros(shape=(T,), dtype=np.float64)
    f_K = np.zeros(shape=(T,), dtype=np.float64)
    ref = np.log(M)
    iterate_fi(log_N_K, log_N_i, f_K, b_K_i, scratch_M, scratch_T, f_i)
    assert_allclose(f_i, ref, atol=1.0E-15)

def test_wham_fi_with_ascending_K():
    T = 5
    M = 10
    N_K_i = np.ones(shape=(T, M), dtype=np.float64)
    log_N_K = np.log(N_K_i.sum(axis=1))
    log_N_i = np.log(N_K_i.sum(axis=0))
    f_i = np.zeros(shape=(M,), dtype=np.float64)
    b_K_i = np.log(np.array([[i*M for i in range(1, M + 1)] for K in range(T)], dtype=np.float64))
    scratch_M = np.zeros(shape=(M,), dtype=np.float64)
    scratch_T = np.zeros(shape=(T,), dtype=np.float64)
    f_K = np.zeros(shape=(T,), dtype=np.float64)
    ref = -np.log((np.arange(M).astype(np.float64) + 1) / (0.5 * M * (M + 1)))
    iterate_fi(log_N_K, log_N_i, f_K, b_K_i, scratch_M, scratch_T, f_i)
    assert_allclose(f_i, ref, atol=1.0E-15)



# This file is part of thermotools.
#
# Copyright 2015 Computational Molecular Biology Group, Freie Universitaet Berlin (GER)
#
# thermotools is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from nose.tools import assert_true
from thermotools.lse import logsumexp, logsumexp_pair
import numpy as np

def test_logsumexp_zeros_short():
    N = 5
    data = np.zeros(shape=(N,), dtype=np.float64)
    assert_true(np.abs(logsumexp(data) - np.log(N)) < 1.0E-15)
    assert_true(np.abs(logsumexp(-data) - np.log(N)) < 1.0E-15)

def test_logsumexp_zeros_long():
    N = 10000
    data = np.zeros(shape=(N,), dtype=np.float64)
    assert_true(np.abs(logsumexp(data) - np.log(N)) < 1.0E-15)
    assert_true(np.abs(logsumexp(-data) - np.log(N)) < 1.0E-15)

def test_logsumexp_converged_geometric_series():
    data = np.arange(10000).astype(np.float64)
    assert_true(np.abs(logsumexp(-data) - 0.45867514538708193) < 1.0E-15)
    
def test_logsumexp_truncated_diverging_gemoetric_series():
    data = np.arange(10000).astype(np.float64)
    assert_true(np.abs(logsumexp(data) - 9999.4586751453862) < 1.0E-15)

def test_logsumexp_pair():
    assert_true(np.abs(logsumexp_pair(0.0, 0.0) - np.log(2.0)) < 1.0E-15)
    assert_true(np.abs(logsumexp_pair(1.0, 1.0) - (1.0 + np.log(2.0))) < 1.0E-15)
    assert_true(np.abs(logsumexp_pair(10.0, 10.0) - (10.0 + np.log(2.0))) < 1.0E-15)
    assert_true(np.abs(logsumexp_pair(100.0, 100.0) - (100.0 + np.log(2.0))) < 1.0E-15)
    assert_true(np.abs(logsumexp_pair(1000.0, 1000.0) - (1000.0 + np.log(2.0))) < 1.0E-15)
    assert_true(np.abs(logsumexp_pair(10.0, 0.0) - 10.000045398899218) < 1.0E-15)
    assert_true(np.abs(logsumexp_pair(0.0, 10.0) - 10.000045398899218) < 1.0E-15)
    assert_true(np.abs(logsumexp_pair(100.0, 0.0) - 100.0) < 1.0E-15)
    assert_true(np.abs(logsumexp_pair(0.0, 100.0) - 100.0) < 1.0E-15)
    assert_true(np.abs(logsumexp_pair(1000.0, 0.0) - 1000.0) < 1.0E-15)
    assert_true(np.abs(logsumexp_pair(0.0, 1000.0) - 1000.0) < 1.0E-15)
